/*
 * $Id: 53b55dfab1c9c21d880a311ba672f749fa1f6dac $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>

#include "ntl.h"
#include "ntlc.h"
#include "ntg.h"
#include "nt_alloc.h"

#define NTL_MAX_TPLC	64

static const char rcsid[] = "$Id: 53b55dfab1c9c21d880a311ba672f749fa1f6dac $";

int ntlc_error = 0;
int ntlc_error_line = 0;
char *ntlc_error_file = NULL;
const char *ntlc_error_str = NULL;

static char *ntlc_fixarg(char *);

static char *
ntlc_fixarg(char *s)
{
	int i;

	for (i = 0; s[i] != '\0'; i++)
		if (s[i] == '_')
			s[i] = ' ';

	return s;
}

void
ntlc_tpl_free(NTL_TPL *tpl)
{
	int i;

	if (tpl == NULL)
		return;

	if (tpl->desc != NULL)
		free(tpl->desc);
	if (tpl->attr != NULL)
		free(tpl->attr);
	if (tpl->regex != NULL)
		free(tpl->regex);
	if (tpl->en != NULL) {
		for (i = 0; tpl->en[i] != NULL; i++)
			free(tpl->en[i]);
		free(tpl->en);
	}
	free(tpl);
	tpl = NULL;
}

void
ntlc_tplc_free(NTL_TPLC *tplc)
{
	int i;

	if (tplc == NULL)
		return;

	if (tplc->name != NULL)
		free(tplc->name);
	if (tplc->oc != NULL) {
		i = 0;
		while (tplc->oc[i] != NULL)
			free(tplc->oc[i++]);
		free(tplc->oc);
	}
	if (tplc->tpl != NULL) {
		i = 0;
		while (tplc->tpl[i] != NULL)
			ntlc_tpl_free(tplc->tpl[i++]);
		free(tplc->tpl);
	}
	if (tplc->crattr != NULL)
		ntl_entry_free(tplc->crattr);
	if (tplc->usrattr != NULL)
		ntl_entry_free(tplc->usrattr);
	free(tplc);
	tplc = NULL;
}

NTL_TPLC *
ntlc_tpl_read(const char *fn)
{
	NTL_TPLC *tplc;
	FILE *fp;
	char rfn[256];
	char b[512];
	char desc[64], attr[64], type[64], misc[256];
	char *r;
	int a, null, multi;
	int line;
	int num;
	int i;

	ntlc_error = 0;
	ntlc_error_line = 0;
	ntlc_error_file = NULL;
	ntlc_error_str = NULL;

	(void)snprintf(rfn, sizeof(rfn), "%s/%s/%s", getenv("HOME"), NT_CFG_DEFAULT_DIR, fn);
	if ((fp = fopen(rfn, "r")) == NULL) {
		ntlc_error = 4;
		ntlc_error_str = "unable to open template file";
		ntlc_error_file = nt_strdup(rfn);
		return NULL;
	}

	tplc = nt_calloc(1, sizeof(NTL_TPLC));
	tplc->name = NULL;
	tplc->oc = nt_calloc(32, sizeof(char *));
	tplc->oc[0] = NULL;
	tplc->ocmax = 32;
	tplc->tpl = NULL;
	tplc->tplmax = 64;
	tplc->crattr = NULL;
	tplc->crmax = 32;
	tplc->usrattr = NULL;
	tplc->usrmax = 32;

	/*
	 * read header
	 */
	i = 0;
	line = 0;
	while ((r = fgets(b, sizeof(b), fp)) != NULL) {
		line++;
		(void)sscanf(b, "%63s %63s", desc, attr);
		if (*b == '\n' || *desc == '#')
			continue;
		else if (strcmp(desc, "TEMPLATE") == 0)
			break;

		else if (strcmp(desc, "name") == 0)
			tplc->name = nt_strdup(attr);
		else if (strcmp(desc, "oc") == 0) {
			if (i < tplc->ocmax - 1) {
				tplc->oc[i] = nt_strdup(attr);
				tplc->oc[++i] = NULL;
			} else {
				ntlc_error = 8;
				ntlc_error_str = "too many ocs";
				goto ntlc_tpl_read_fail;
			}
		} else {
			ntlc_error = 6;
			ntlc_error_str = "unknown option in template";
			goto ntlc_tpl_read_fail;
		}
	}
	if (r == NULL) {
		ntlc_error = 7;
		ntlc_error_str = "missing template part";
		goto ntlc_tpl_read_fail;
	}

	/*
	 * read real template
	 */
	i = 0;
	tplc->tpl = nt_calloc((size_t)tplc->tplmax, sizeof(NTL_TPL *));
	tplc->tpl[0] = NULL;
	while ((r = fgets(b, sizeof(b), fp)) != NULL) {
		line++;
		num = sscanf(b, "%63s %63s %i %63s %i %i %255[^]", desc, attr, &a, type, &null, &multi, misc);
		if (*b == '\n' || *desc == '#')
			continue;
		else if (strcmp(desc, "CREATE") == 0)
			break;

		else if (*desc != '-' && num != 7) {
			ntlc_error = 13;
			ntlc_error_str = "wrong number of values for template line";
			goto ntlc_tpl_read_fail;
		} else if (i < tplc->tplmax - 1) {
			tplc->tpl[i] = nt_calloc(1, sizeof(NTL_TPL));
			tplc->tpl[i+1] = NULL;
			tplc->tpl[i]->type = 0;
			tplc->tpl[i]->prec = 0;
			tplc->tpl[i]->regex = NULL;
			tplc->tpl[i]->en = NULL;
			if (*desc == '-') {
				tplc->tpl[i]->desc = nt_strdup("");
				tplc->tpl[i]->attr = nt_strdup("");
			} else {
				tplc->tpl[i]->desc = ntlc_fixarg(strdup(desc));
				tplc->tpl[i]->attr = nt_strdup(attr);
				tplc->tpl[i]->a = a;
				if (strcmp(type, "string") == 0)
					tplc->tpl[i]->type = 0;
				else if (strcmp(type, "none") == 0)
					tplc->tpl[i]->type = -1;
				else if (strcmp(type, "mail") == 0)
					tplc->tpl[i]->type = NTL_T_MAIL;
				else if (strcmp(type, "name") == 0)
					tplc->tpl[i]->type = NTL_T_NAME;
				else if (strcmp(type, "cn") == 0)
					tplc->tpl[i]->type = NTL_T_CN;
				else if (strcmp(type, "int") == 0)
					tplc->tpl[i]->type = NTL_T_INT;
				else if (strcmp(type, "num") == 0) {
					tplc->tpl[i]->type = NTL_T_NUM;
					tplc->tpl[i]->prec = atoi(misc);
				} else if (strcmp(type, "bool") == 0)
					tplc->tpl[i]->type = NTL_T_BOOL;
				else if (strcmp(type, "alpha") == 0)
					tplc->tpl[i]->type = NTL_T_ALPHA;
				else if (strcmp(type, "alnum") == 0)
					tplc->tpl[i]->type = NTL_T_ALNUM;
				else if (strcmp(type, "ipv4") == 0)
					tplc->tpl[i]->type = NTL_T_IPV4;
				else if (strcmp(type, "regex") == 0) {
					regex_t regex;
					int e;

					tplc->tpl[i]->type = NTL_T_REGEX;
					if (strlen(misc) > 0)
						misc[strlen(misc) - 1] = '\0';
					nt_asprintf(&tplc->tpl[i]->regex, "^( |%s) *$", misc);
					if ((e = regcomp(&regex, tplc->tpl[i]->regex, REG_EXTENDED)) != 0) {
						ntlc_error = 16;
						ntlc_error_str = "unable to compile regex";
						ntlc_error_file = nt_calloc(512, sizeof(char));
						(void)regerror(e, &regex, ntlc_error_file, 512);
						goto ntlc_tpl_read_fail2;
					}
					regfree(&regex);
				} else if (strcmp(type, "enum") == 0) {
					char *p, *last;
					int j;

					tplc->tpl[i]->type = NTL_T_ENUM;
					tplc->tpl[i]->en = nt_calloc(64, sizeof(char *));
					if (strlen(misc) > 0)
						misc[strlen(misc) - 1] = '\0';
					for (j = 0, p = strtok_r(misc, "|", &last); j < 63 && p != NULL; (p = strtok_r(NULL, "|", &last)), j++)
						tplc->tpl[i]->en[j] = nt_strdup(p);
					tplc->tpl[i]->en[j] = NULL;

					if (p != NULL) {
						ntlc_error = 17;
						ntlc_error_str = "too many tokens in enum (max. 63)";
						goto ntlc_tpl_read_fail;
					}
				} else {
					ntlc_error = 15;
					ntlc_error_str = "unknown field type";
					goto ntlc_tpl_read_fail;
				}
				tplc->tpl[i]->null = null;
				tplc->tpl[i]->multi = multi;
			}
			i++;
		} else {
			ntlc_error = 9;
			ntlc_error_str = "too many attributes in template";
			goto ntlc_tpl_read_fail;
		}
	}
	if (r == NULL) {
		ntlc_error = 10;
		ntlc_error_str = "missing creation part";
		goto ntlc_tpl_read_fail;
	}

	/*
	 * read (attr, val) pairs for creation of a new object for this template
	 */
	i = 0;
	tplc->crattr = nt_calloc((size_t)tplc->crmax, sizeof(NTL_ATTR));
	tplc->crattr[0].name = NULL;
	tplc->crattr[0].val = NULL;
	tplc->crattr[0].tsize = tplc->crmax;
	while ((r = fgets(b, sizeof(b), fp)) != NULL) {
		line++;
		num = sscanf(b, "%63s %63s", desc, attr);
		if (*b == '\n' || *desc == '#')
			continue;
		else if (strcmp(desc, "USER") == 0)
			break;

		else if (num != 2) {
			ntlc_error = 14;
			ntlc_error_str = "wrong number of values for create template line";
			goto ntlc_tpl_read_fail;
		} else if (i < tplc->crmax - 1)
			ntl_attr_add(tplc->crattr, desc, ntlc_fixarg(attr));
		else {
			ntlc_error = 11;
			ntlc_error_str = "too many attributs in creation list";
			goto ntlc_tpl_read_fail;
		}
	}
	if (r == NULL) {
		ntlc_error = 10;
		ntlc_error_str = "missing user part";
		goto ntlc_tpl_read_fail;
	}

	/*
	 * read further needed attributes (to be user specified) for
	 * creation of a new object for this template
	 */
	i = 0;
	tplc->usrattr = nt_calloc((size_t)tplc->usrmax, sizeof(NTL_ATTR));
	tplc->usrattr[0].name = NULL;
	tplc->usrattr[0].val = NULL;
	tplc->usrattr[0].tsize = tplc->usrmax;
	while (fgets(b, sizeof(b), fp) != NULL) {
		line++;
		num = sscanf(b, "%63s %63s", desc, attr);
		if (*b == '\n' || *desc == '#')
			continue;

		else if (num != 2) {
			ntlc_error = 15;
			ntlc_error_str = "wrong number of values for user template line";
			goto ntlc_tpl_read_fail;
		} else if (i < tplc->usrmax - 1)
			ntl_attr_add_nocheck(tplc->usrattr, desc, ntlc_fixarg(attr));
		else {
			ntlc_error = 12;
			ntlc_error_str = "too many attributs in user list";
			goto ntlc_tpl_read_fail;
		}
	}

	(void)fclose(fp);
	return tplc;

ntlc_tpl_read_fail:
	ntlc_error_file = nt_strdup(rfn);
ntlc_tpl_read_fail2:
	ntlc_error_line = line;
	ntlc_tplc_free(tplc);
	(void)fclose(fp);
	return NULL;
}

NTL_CFG *
ntlc_cfg_read(const char *fn, const char *profile)
{
	NTL_CFG *cfg;
	FILE *fp;
	char rfn[256];
	char *rprofile;
	char b[512];
	char obuf[128];
	char abuf[256];
	char *p;
	int tplcount;
	int line;

	ntlc_error = 0;
	ntlc_error_line = 0;
	ntlc_error_file = NULL;
	ntlc_error_str = NULL;

	if (fn == NULL)
		(void)snprintf(rfn, sizeof(rfn), "%s/%s", getenv("HOME"), NT_CFG_DEFAULT);
	else
		(void)snprintf(rfn, sizeof(rfn), "%s", fn);

	cfg = nt_calloc(1, sizeof(NTL_CFG));
	cfg->dn = cfg->cred = cfg->basedn = cfg->host = NULL;
	cfg->tplc = nt_calloc(NTL_MAX_TPLC, sizeof(NTL_TPLC *));
	cfg->tplc[0] = NULL;
	cfg->port = 0;
	cfg->secure = 0;
	cfg->showop = 0;

	if ((fp = fopen(rfn, "r")) == NULL) {
		ntlc_error = 1;
		ntlc_error_str = "unable to open config file";
		ntlc_error_file = nt_strdup(rfn);
		ntlc_cfg_free(cfg);
		return NULL;
	}

	/*
	 * skip forward until "profiles" section
	 */
	line = 1;
	while ((p = fgets(b, sizeof(b), fp)) != NULL && strncmp(b, "[profiles]", strlen("[profiles]")) != 0)
		line++;

	if (p == NULL) {
		ntlc_error = 20;
		ntlc_error_str = "unable to find section \"profiles\" in config file";
		ntlc_error_file = nt_strdup(rfn);
		(void)fclose(fp);
		ntlc_cfg_free(cfg);
		return NULL;
	}

	/*
	 * skip forward until we reach wanted profile, or "default", if profile == NULL
	 */
	if (profile == NULL)
		rprofile = nt_strdup("=default=");
	else
		nt_asprintf(&rprofile, "=%s=", profile);

	while ((p = fgets(b, sizeof(b), fp)) != NULL && *b != '[' && strncmp(b, rprofile, strlen(rprofile)) != 0)
		line++;
	line++;

	if (p == NULL || *b == '[') {
		ntlc_error = 21;
		ntlc_error_str = "unable to find specified profile in config file";
		nt_asprintf(&ntlc_error_file, "%s %s", rfn, rprofile);
		(void)fclose(fp);
		ntlc_cfg_free(cfg);
		free(rprofile);
		return NULL;
	}
	free(rprofile);

	tplcount = 0;
	while (fgets(b, sizeof(b), fp) != NULL && *b != '[' && *b != '=') {
		line++;
		(void)sscanf(b, "%127s %255s", obuf, abuf);
		if (*b == '\n')
			continue;

		if (strcmp(obuf, "dn") == 0)
			cfg->dn = nt_strdup(ntlc_fixarg(abuf));
		else if (strcmp(obuf, "cred") == 0)
			cfg->cred = nt_strdup(abuf);
		else if (strcmp(obuf, "basedn") == 0)
			cfg->basedn = nt_strdup(ntlc_fixarg(abuf));
		else if (strcmp(obuf, "host") == 0)
			cfg->host = nt_strdup(abuf);
		else if (strcmp(obuf, "port") == 0)
			cfg->port = atol(abuf);
		else if (strcmp(obuf, "secure") == 0) {
			if (strcmp(abuf, "yes") == 0)
				cfg->secure = 1;
			else if (strcmp(abuf, "no") == 0)
				cfg->secure = 0;
			else {
				ntlc_error = 3;
				ntlc_error_str = "invalid arg for secure";
				goto ntlc_cfg_read_fail;
			}
		} else if (strcmp(obuf, "template") == 0) {
			if (tplcount < NTL_MAX_TPLC - 1) {
				if ((cfg->tplc[tplcount] = ntlc_tpl_read(abuf)) == NULL)
					goto ntlc_cfg_read_fail2;
				cfg->tplc[++tplcount] = NULL;
			} else {
				ntlc_error = 5;
				ntlc_error_str = "too many templates";
				goto ntlc_cfg_read_fail;
			}
		} else if (strcmp(obuf, "show_operational") == 0) {
			if (strcmp(abuf, "yes") == 0)
				cfg->showop = 1;
			else if (strcmp(abuf, "no") == 0)
				cfg->showop = 0;
			else {
				ntlc_error = 30;
				ntlc_error_str = "invalid arg for show_operational";
				goto ntlc_cfg_read_fail;
			}
		} else if (*obuf != '#') {
			ntlc_error = 2;
			ntlc_error_str = "unknown option in section \"profiles\" of config file";
			goto ntlc_cfg_read_fail;
		}
	}

	if (cfg->host == NULL) {
		ntlc_error = 40;
		ntlc_error_str = "no host to connect to given for profile";
		goto ntlc_cfg_read_fail;
	}

	if (cfg->basedn == NULL)
		cfg->basedn = nt_strdup("");
	if (cfg->dn == NULL)
		cfg->dn = nt_strdup("");
	if (cfg->cred == NULL)
		cfg->cred = nt_strdup("");

	(void)fclose(fp);
	return cfg;

ntlc_cfg_read_fail:
	ntlc_error_line = line;
	ntlc_error_file = nt_strdup(rfn);
ntlc_cfg_read_fail2:
	ntlc_cfg_free(cfg);
	(void)fclose(fp);
	return NULL;
}

void
ntlc_cfg_free(NTL_CFG *cfg)
{
	int i;

	if (cfg == NULL)
		return;

	if (cfg->dn != NULL)
		free(cfg->dn);
	if (cfg->cred != NULL)
		free(cfg->cred);
	if (cfg->basedn != NULL)
		free(cfg->basedn);
	if (cfg->host != NULL)
		free(cfg->host);
	if (cfg->tplc != NULL) {
		i = 0;
		while (cfg->tplc[i] != NULL)
			ntlc_tplc_free(cfg->tplc[i++]);
		free(cfg->tplc);
	}
	free(cfg);
}

void
ntlc_perror(const char *s)
{

	if (s != NULL)
		(void)fprintf(stderr, "%s: ", s);

	(void)fprintf(stderr, "error reading LDAP config: (%d) %s (%s, line #%d)\n",
	    ntlc_error, ntlc_error_str, (ntlc_error_file != NULL) ? ntlc_error_file : "(null)", ntlc_error_line);

	if (ntlc_error_file != NULL)
		free(ntlc_error_file);
}
