/*
 * $Id: 4bd717bbd69272d455084440027fe45de0d9eb87 $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ldap.h"
#include "ldap_schema.h"

#include "ntl.h"
#include "ntlc.h"
#include "ntls.h"
#include "nt_alloc.h"

#define NTLS_MAX_OC_PER_CLASS	128
#define NTLS_MAX_AT		2048
#define NTLS_ACC_MAX		128
#define NTLS_INTERPOLATE_MAX	256
#define NTLS_SOI_PREFIX		"1.3.6.1.4.1.1466.115.121.1."

static int ntls_acc_merge(const NTLS_SCHEMA *, char **, const char **);

static const char rcsid[] = "$Id: 4bd717bbd69272d455084440027fe45de0d9eb87 $";

int ntls_error = 0;
const char *ntls_error_str = NULL;

NTLS_SCHEMA *
ntls_read_schema(LDAP *ld)
{
	LDAPMessage *res, *entry;
	NTL_ATTR *attr;
	NTLS_SCHEMA *schema;
	const char *a_subschemaSubentry[] = { "subschemaSubentry", NULL };
	const char *a_all[] = { "objectClasses", "attributeTypes", NULL };
	char *subschema;

	res = entry = NULL;
	attr = NULL;
	subschema = NULL;
	ntls_error = 0;

	schema = nt_calloc(1, sizeof(NTLS_SCHEMA));
	schema->ocs = NULL;
	schema->ats = NULL;

	/*
	 * Step 1:
	 * Get subschemaSubentry attribute from Root.
	 */
	if ((res = ntl_search(ld, "", LDAP_SCOPE_BASE, NULL, a_subschemaSubentry)) == NULL) {
		ntls_error = 1;
		ntls_error_str = "Unable to get root entry (1).";
	} else if ((entry = ldap_first_entry(ld, res)) == NULL) {
		ntls_error = 2;
		ntls_error_str = "Unable to get root entry (2).";
	} else if ((attr = ntl_entry(ld, entry)) == NULL) {
		ntls_error = 3;
		ntls_error_str = "Unable to get root entry (3).";
	} else if (*ntl_lookup(attr, "subschemaSubentry") == NULL) {
		ntls_error = 4;
		ntls_error_str = "Unable to get subschemaSubentry attribute.\n";
	} else
		subschema = nt_strdup(*ntl_lookup(attr, "subschemaSubentry"));

	if (attr != NULL)
		ntl_entry_free(attr);

	if (ntls_error != 0)
		goto ntls_schema_read_out;

	/*
	 * Step 2:
	 * Get Subschema Entry.
	 */
	if (res != NULL)
		(void)ldap_msgfree(res);
	if ((res = ntl_search(ld, subschema, LDAP_SCOPE_BASE, NULL, a_all)) == NULL) {
		ntls_error = 5;
		ntls_error_str = "Unable to get Subschema entry (1).";
	} else if ((entry = ldap_first_entry(ld, res)) == NULL) {
		ntls_error = 6;
		ntls_error_str = "Unable to get Subschema entry (2).";
	} else if ((attr = ntl_entry(ld, entry)) == NULL) {
		ntls_error = 7;
		ntls_error_str = "Unable to get Subschema entry (3).";
		ntl_entry_free(attr);
	}
	free(subschema);

	if (ntls_error != 0)
		goto ntls_schema_read_out;

	/*
	 * Step 3:
	 * Retrieve all objectClasses and categorize them into a meta structure.
	 */
	{
		LDAPObjectClass *oc, **cur;
		const char *str;
		const char **p;
		int i;
		int e;
		int k_abs, k_str, k_aux;
		int *k;
		int k_all;

		if (*(p = ntl_lookup(attr, "objectClasses")) == NULL) {
			ntls_error = 10;
			ntls_error_str = "Unable to get attribute objectClasses from Subschema.";
			goto ntls_schema_read_out;
		}

		schema->ocs = nt_calloc(1, sizeof(NTLS_OC));
		schema->ocs->abs = nt_calloc(NTLS_MAX_OC_PER_CLASS, sizeof(LDAPObjectClass *));
		schema->ocs->str = nt_calloc(NTLS_MAX_OC_PER_CLASS, sizeof(LDAPObjectClass *));
		schema->ocs->aux = nt_calloc(NTLS_MAX_OC_PER_CLASS, sizeof(LDAPObjectClass *));
		schema->ocs->all = nt_calloc(NTLS_MAX_OC_PER_CLASS * 3, sizeof(LDAPObjectClass *));
		schema->ocs->abs[0] = schema->ocs->str[0] = schema->ocs->aux[0] = schema->ocs->all[0] = NULL;

		k = NULL;
		k_abs = k_str = k_aux = k_all = 0;
		for (i = 0; p[i] != NULL; i++) {
			if ((oc = ldap_str2objectclass(p[i], &e, &str, LDAP_SCHEMA_ALLOW_ALL)) != NULL) {
				switch (oc->oc_kind) {
				default:
				case LDAP_SCHEMA_ABSTRACT:
					cur = schema->ocs->abs;
					k = &k_abs;
					break;
				case LDAP_SCHEMA_STRUCTURAL:
					cur = schema->ocs->str;
					k = &k_str;
					break;
				case LDAP_SCHEMA_AUXILIARY:
					cur = schema->ocs->aux;
					k = &k_aux;
					break;
				}

				if (*k < NTLS_MAX_OC_PER_CLASS - 1) {
					cur[*k] = oc;
					cur[*k+1] = NULL;
					schema->ocs->all[k_all] = oc;
					schema->ocs->all[k_all+1] = NULL;
					(*k)++;
					k_all++;
				} else {
					ntls_error = 12;
					ntls_error_str = "Too many objectClasses, increase NTLS_MAX_OC_PER_CLASS.";
					ldap_objectclass_free(oc);
					goto ntls_schema_read_out;
				}
			} else {
				ntls_error = 11;
				ntls_error_str = "Unable to parse an objectClass from Subschema.";
				goto ntls_schema_read_out;
			}
		}
	}

	/*
	 * Step 4:
	 * Retrieve all attributeTypes and categorize them into a meta structure.
	 */
	{
		LDAPAttributeType *cur;
		const char **p;
		const char *str;
		int e;
		int i;

		if (*(p = ntl_lookup(attr, "attributeTypes")) == NULL) {
			ntls_error = 10;
			ntls_error_str = "Unable to get attribute attributeTypes from Subschema.";
			goto ntls_schema_read_out;
		}

		schema->ats = nt_calloc(NTLS_MAX_AT, sizeof(LDAPAttributeType *));
		schema->ats[0] = NULL;

		for (i = 0; p[i] != NULL; i++) {
			if ((cur = ldap_str2attributetype(p[i], &e, &str, LDAP_SCHEMA_ALLOW_ALL)) == NULL) {
				ntls_error = 11;
				ntls_error_str = "Unable to parse an attributeType from Subschema.";
				goto ntls_schema_read_out;
			}

			if (i < NTLS_MAX_AT - 1) {
				schema->ats[i] = cur;
				schema->ats[i+1] = NULL;
			} else {
				ntls_error = 12;
				ntls_error_str = "Too many attributes, increase NTLS_MAX_AT.";
				ldap_attributetype_free(cur);
				goto ntls_schema_read_out;
			}
		}
	}

ntls_schema_read_out:
	if (ntls_error != 0) {
		ntls_schema_free(schema); /* note: nlt_schema_free recognizes NULL-pointer in members */
		schema = NULL;
	}
	if (attr != NULL)
		ntl_entry_free(attr);
	if (res != NULL)
		(void)ldap_msgfree(res);

	return schema;
}

void
ntls_schema_free(NTLS_SCHEMA *schema)
{
	int i;

	if (schema == NULL)
		return;

	if (schema->ocs != NULL) {
		if (schema->ocs->abs != NULL) {
			for (i = 0; schema->ocs->abs[i] != NULL; i++)
				ldap_objectclass_free(schema->ocs->abs[i]);
			free(schema->ocs->abs);
		}
		if (schema->ocs->str != NULL) {
			for (i = 0; schema->ocs->str[i] != NULL; i++)
				ldap_objectclass_free(schema->ocs->str[i]);
			free(schema->ocs->str);
		}
		if (schema->ocs->aux != NULL) {
			for (i = 0; schema->ocs->aux[i] != NULL; i++)
				ldap_objectclass_free(schema->ocs->aux[i]);
			free(schema->ocs->aux);
		}
		if (schema->ocs->all != NULL)
			free(schema->ocs->all);
		free(schema->ocs);
	}

	if (schema->ats != NULL) {
		for (i = 0; schema->ats[i] != NULL; i++)
			ldap_attributetype_free(schema->ats[i]);
		free(schema->ats);
	}

	free(schema);
}

const LDAPAttributeType *
ntls_get_at_from_oid(const LDAPAttributeType **ats, const char *oid)
{
	int i;

	if (ats == NULL) {
		ntls_error = 30;
		ntls_error_str = "AttributeTypes empty.";
		return NULL;
	}

	for (i = 0; ats[i] != NULL; i++)
		if (ats[i]->at_oid != NULL && strcasecmp(ats[i]->at_oid, oid) == 0)
			return ats[i];

	ntls_error = 31;
	ntls_error_str = "AttributeType not found.";
	return NULL;
}

const LDAPObjectClass *
ntls_get_oc_from_oid(const LDAPObjectClass **ocs, const char *oid)
{
	int i;

	if (ocs == NULL) {
		ntls_error = 32;
		ntls_error_str = "ObjectClasses empty.";
		return NULL;
	}

	for (i = 0; ocs[i] != NULL; i++)
		if (ocs[i]->oc_oid != NULL && strcasecmp(ocs[i]->oc_oid, oid) == 0)
			return ocs[i];

	ntls_error = 33;
	ntls_error_str = "ObjectClass not found.";
	return NULL;
}

const LDAPAttributeType *
ntls_get_at_from_name(const LDAPAttributeType **ats, const char *name)
{
	int i, j;

	if (ats == NULL) {
		ntls_error = 30;
		ntls_error_str = "AttributeTypes empty.";
		return NULL;
	}

	for (i = 0; ats[i] != NULL; i++)
		for (j = 0; ats[i]->at_names[j] != NULL; j++)
			if (strcasecmp(ats[i]->at_names[j], name) == 0)
				return ats[i];

	ntls_error = 31;
	ntls_error_str = "AttributeType not found.";
	return NULL;
}

const LDAPObjectClass *
ntls_get_oc_from_name(const LDAPObjectClass **ocs, const char *name)
{
	int i, j;

	if (ocs == NULL) {
		ntls_error = 32;
		ntls_error_str = "ObjectClasses empty.";
		return NULL;
	}

	for (i = 0; ocs[i] != NULL; i++)
		for (j = 0; ocs[i]->oc_names[j] != NULL; j++)
			if (strcasecmp(ocs[i]->oc_names[j], name) == 0)
				return ocs[i];

	ntls_error = 33;
	ntls_error_str = "ObjectClass not found.";
	return NULL;
}

/**
 * Accumlate attributes:
 * 'r' == required
 * 'a' == auxiliary
 * '*' == all
 */
char **
ntls_acc_ats(const NTLS_SCHEMA *schema, const char **ocs, int type)
{
	const LDAPObjectClass *cur_oc;
	char **acc;
	int i;

	acc = nt_calloc(NTLS_ACC_MAX, sizeof(char *));
	acc[0] = NULL;

	for (i = 0; ocs[i] != NULL; i++) {
		if ((cur_oc = ntls_get_oc_from_name((const LDAPObjectClass **) schema->ocs->all, ocs[i])) == NULL) {
			ntls_error = 40;
			ntls_error_str = "ObjectClass not found.";
			ntls_acc_free(acc);
			return NULL;
		}

		if (type == 'r' || type == '*')
			if (ntls_acc_merge(schema, acc, (const char **)cur_oc->oc_at_oids_must) != 0) {
				ntls_acc_free(acc);
				return NULL;
			}
		if (type == 'a' || type == '*')
			if (ntls_acc_merge(schema, acc, (const char **)cur_oc->oc_at_oids_may) != 0) {
				ntls_acc_free(acc);
				return NULL;
			}
	}

	return acc;
}

/**
 * Merge names(oids) into acc.
 * Note:  We're supposed to get OIDs in oc->**oc_at_oids_may, but we get names actually.
 *        So we look for both below.
 */
static int
ntls_acc_merge(const NTLS_SCHEMA *schema, char **acc, const char **oids)
{
	const LDAPAttributeType *cur_at;
	int i, j;

	if (oids == NULL)
		return 0;

	for (i = 0; oids[i] != NULL; i++) {
		if ((cur_at = ntls_get_at_from_oid((const LDAPAttributeType **) schema->ats, oids[i])) != NULL
		    || (cur_at = ntls_get_at_from_name((const LDAPAttributeType **) schema->ats, oids[i])) != NULL) {
			/* Already in acc? */
			for (j = 0; acc[j] != NULL && j < NTLS_ACC_MAX - 1; j++)
				if (strcasecmp(cur_at->at_names[0], acc[j]) == 0)
					break;
			if (acc[j] == NULL && j >= NTLS_ACC_MAX - 2) {
				ntls_error = 41;
				ntls_error_str = "Too many attributes.";
				return 41;
			} else if (acc[j] == NULL) {
				acc[j] = nt_strdup(cur_at->at_names[0]);
				acc[j+1] = NULL;
			}
		} else if (strcasecmp(oids[i], "x500uniqueIdentifier") == 0)
			; /* ignore */
		else {
			ntls_error = 42;
			ntls_error_str = "Attribute from ObjectClass not found.";
			return 42;
		}
	}
	return 0;
}

void
ntls_acc_free(char **acc)
{
	int i;

	if (acc == NULL)
		return;

	for (i = 0; acc[i] != NULL; i++)
		free(acc[i]);
	free(acc);
}

void
ntls_ocs_free(char **ocs)
{
	ntls_acc_free(ocs);
}

char **
ntls_interpolate_ocs(const NTLS_SCHEMA *schema, const char **ocs)
{
	const LDAPObjectClass *cur_oc, *sup;
	char **i_ocs;
	int flag;
	int i, j, k, l;

	i_ocs = nt_calloc(NTLS_INTERPOLATE_MAX, sizeof(char *));
	i_ocs[0] = NULL;

	for (i = 0; ocs[i] != NULL; i++) {
		/*
		 * Step 1:
		 * Fit in current OC, if not already in.
		 */
		for (j = 0; i_ocs[j] != NULL && j < NTLS_INTERPOLATE_MAX - 1; j++)
			if (strcasecmp(ocs[i], i_ocs[j]) == 0)
				break;
		if (i_ocs[j] == NULL && j >= NTLS_INTERPOLATE_MAX - 2) {
			ntls_error = 50;
			ntls_error_str = "Too many interpolated ObjectClasses.";
			goto ntls_interpolate_ocs_out;
		} else if (i_ocs[j] == NULL) {
			i_ocs[j] = nt_strdup(ocs[i]);
			i_ocs[j+1] = NULL;
		}

		/*
		 * Step 2:
		 * Fit in superior OCs of current OC.
		 * Note:  We're supposed to get OIDs in cur_oc->**oc_sup_oids, but we get names actually.
		 *        So we look for both below.
		 */
		if ((cur_oc = ntls_get_oc_from_name((const LDAPObjectClass **) schema->ocs->all, ocs[i])) == NULL) {
			ntls_error = 51;
			ntls_error_str = "Could not find ObjectClass.";
			goto ntls_interpolate_ocs_out;
		}
		if (cur_oc->oc_sup_oids == NULL)
			continue; /* ignore */
		for (j = 0; cur_oc->oc_sup_oids[j] != NULL; j++) {
			if ((sup = ntls_get_oc_from_oid((const LDAPObjectClass **) schema->ocs->all, cur_oc->oc_sup_oids[j])) == NULL
			    && (sup = ntls_get_oc_from_name((const LDAPObjectClass **) schema->ocs->all, cur_oc->oc_sup_oids[j])) == NULL) {
				ntls_error = 52;
				ntls_error_str = "Could not find specified superior ObjectClass.";
				goto ntls_interpolate_ocs_out;
			}
			for (k = 0; i_ocs[k] != NULL && k < NTLS_INTERPOLATE_MAX - 1; k++)
				if (strcasecmp(i_ocs[k], sup->oc_names[0]) == 0)
					break;
			if (i_ocs[k] == NULL && k >= NTLS_INTERPOLATE_MAX - 2) {
				ntls_error = 50;
				ntls_error_str = "Too many interpolated ObjectClasses.";
				goto ntls_interpolate_ocs_out;
			} else if (i_ocs[k] == NULL) {
				i_ocs[k] = nt_strdup(sup->oc_names[0]);
				i_ocs[k+1] = NULL;
			}
		}

		/*
		 * Step 3:
		 * Fit in superior OC (if any) of all OCs in list.
		 * Do this as long as we find new ones.
		 */
		flag = 1;
		while (flag) {
			flag = 0;
			for (j = 0; i_ocs[j] != NULL && j < NTLS_INTERPOLATE_MAX - 1; j++) {
				if ((cur_oc = ntls_get_oc_from_name((const LDAPObjectClass **) schema->ocs->all, i_ocs[j])) == NULL) {
					ntls_error = 53;
					ntls_error_str = "Internal Error:  Could not find an ObjectClass again in step 3.";
					goto ntls_interpolate_ocs_out;
				}
				if (cur_oc->oc_sup_oids == NULL)
					continue; /* ignore */
				for (k = 0; cur_oc->oc_sup_oids[k] != NULL; k++) {
					if ((sup = ntls_get_oc_from_oid((const LDAPObjectClass **) schema->ocs->all, cur_oc->oc_sup_oids[k])) == NULL
					    && (sup = ntls_get_oc_from_name((const LDAPObjectClass **) schema->ocs->all, cur_oc->oc_sup_oids[k])) == NULL) {
						ntls_error = 54;
						ntls_error_str = "Could not find specified superior ObjectClass in step 3.";
						goto ntls_interpolate_ocs_out;
					}
					for (l = 0; i_ocs[l] != NULL && l < NTLS_INTERPOLATE_MAX - 1; l++)
						if (strcasecmp(i_ocs[l], sup->oc_names[0]) == 0)
							break;
					if (i_ocs[l] == NULL && l >= NTLS_INTERPOLATE_MAX - 2) {
						ntls_error = 50;
						ntls_error_str = "Too many interpolated ObjectClasses.";
						goto ntls_interpolate_ocs_out;
					} else if (i_ocs[l] == NULL) {
						i_ocs[l] = nt_strdup(sup->oc_names[0]);
						i_ocs[l+1] = NULL;
						flag = 1;
					}
				}
			}
		}
	}
	ntls_error = 0;
ntls_interpolate_ocs_out:
	if (ntls_error) {
		ntls_ocs_free(i_ocs);
		return NULL;
	}
	return i_ocs;
}

/**
 * Determine, if an AT is human readable (according to RfC2252 4.3.2).
 *
 * -1 => don't know
 * 0  => not h-r
 * 1  => h-r
 */
int
ntls_at_human(const LDAPAttributeType *at)
{
	char *s;
	int v;

	if (at == NULL)
		return 1;

	if ((s = at->at_syntax_oid) == NULL)
		return 1;
	else if (strcmp("1.3.6.1.1.16.1", at->at_syntax_oid) == 0)
		return 1;
	else if (strncmp(NTLS_SOI_PREFIX, at->at_syntax_oid, strlen(NTLS_SOI_PREFIX)) != 0)
		return -1;

	s += strlen(NTLS_SOI_PREFIX);
	v = atol(s);

	if (v == 2 || v == 3 || v == 6 || (v >= 7 && v <= 58 && (v != 23 && v != 28 && v != 49)))
		return 1;
	return 0;
}
