/*
 * $Id: 44daa3191a4c33d089746688fa7c32854c283a2e $
 *
 * This file is part of ntldap.
 *
 * Copyright (c) 2006, 2007, 2008 Niklas Olmes <niklas@noxa.de>
 *                                http://www.nevm.de/ntldap/
 *                                NEVM-RIPE
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "ntm.h"
#include "ntu.h"
#include "nt_alloc.h"
#include "base64.h"

#define NTM_SEND_LITERAL(s)	if (fwrite(s, strlen(s), 1, p) != 1) goto ntm_send_error;
#define NTM_SEND_WRITE(b, s)	if (fwrite(b, s, 1, p) != 1) goto ntm_send_error;
#define NTM_PART_MAX_SIZE	10485760UL	/* 10 MB */

static const char rcsid[] = "$Id: 44daa3191a4c33d089746688fa7c32854c283a2e $";

static const char ntm_mimehdr[] = "Mime-Version: 1.0\nContent-Type: text/plain; charset=iso-8859-1\nContent-Disposition: inline\nContent-Transfer-Encoding: 8bit\n";
static const char ntm_mimehdr_multi[] = "Mime-Version: 1.0\nContent-Type: multipart/mixed; boundary=\"ntm/___%lu___/mtn\"\nContent-Disposition: inline\n";
static const char ntm_mimehdr_text[] = "Content-Type: text/plain; charset=iso-8859-1\nContent-Transfer-Encoding: 8bit\nContent-Disposition: inline\n";
static const char ntm_mimehdr_boundary[] = "\n\n--ntm/___%lu___/mtn\n";
static const char ntm_mimehdr_part[] = "Content-Type: %s\nContent-Disposition: attachment; filename=\"%s\"\nContent-Transfer-Encoding: base64\n\n";

char ntm_messageid[128] = "";

int
ntm_send(const NTM_MAIL *m)
{
	FILE *p;
	char buf[65536];
	char multi[512];
	char boundary[128];
	char host[128];
	char domain[128];
	char **cur;
	char *t;
	long rnd;
	int k, i;
	int e;

	if (m->from == NULL || m->to == NULL || m->subject == NULL || m->text == NULL)
		return -2;

#ifdef TEST
	if ((p = popen("cat", "w")) == NULL)
#else
	if ((p = popen("sendmail -it", "r+")) == NULL)
#endif
		return -1;

	*host = *domain = '\0';
	(void)gethostname(host, sizeof(host));
	host[sizeof(host) - 1] = '\0';
	(void)getdomainname(domain, sizeof(domain));
	domain[sizeof(domain) - 1] = '\0';

	if (strchr(host, '.') != NULL)
		*domain = '\0'; /* hostname seems to contain domain, don't use domain */

	ntu_seed();
	rnd = random();
	t = ntu_midtime(time(NULL));
	(void)snprintf(boundary, sizeof(boundary), ntm_mimehdr_boundary, rnd);
	(void)snprintf(multi, sizeof(multi), ntm_mimehdr_multi, rnd);
	(void)snprintf(ntm_messageid, sizeof(ntm_messageid), "Message-Id: <%s.%lu.ntm@%s%s%s>\n", t, random(), host, *domain != '\0' ? "." : "", domain);
	free(t);

	NTM_SEND_LITERAL(ntm_messageid);

	(void)snprintf(buf, sizeof(buf), "From: %s\n", m->from);
	NTM_SEND_WRITE(buf, strlen(buf));

	for (k = 0; k < 3; k++) {
		if (k == 0) {
			NTM_SEND_LITERAL("To: ");
			cur = m->to;
		} else if (k == 1) {
			if (m->cc == NULL)
				continue;
			NTM_SEND_LITERAL("Cc: ");
			cur = m->cc;
		} else { /* k == 2 */
			if (m->bcc == NULL)
				continue;
			NTM_SEND_LITERAL("Bcc: ");
			cur = m->bcc;
		}

		for (i = 0; cur[i] != NULL; i++) {
			(void)snprintf(buf, sizeof(buf), "%s%s", i == 0 ? "" : ", ", cur[i]);
			NTM_SEND_WRITE(buf, strlen(buf));
		}

		NTM_SEND_LITERAL("\n");
	}

	(void)snprintf(buf, sizeof(buf), "Subject: %s\n", m->subject);
	NTM_SEND_WRITE(buf, strlen(buf));

	if (m->parts == NULL) {
		NTM_SEND_LITERAL(ntm_mimehdr);
	} else {
		NTM_SEND_LITERAL(multi);
		NTM_SEND_LITERAL(boundary);
		NTM_SEND_LITERAL(ntm_mimehdr_text);
	}

	NTM_SEND_LITERAL("\n");
	NTM_SEND_LITERAL(m->text);

	if (m->parts == NULL)
		goto ntm_send_done;

	for (i = 0; m->parts[i] != NULL; i++) {
		NTM_SEND_LITERAL(boundary);
		(void)snprintf(buf, sizeof(buf), ntm_mimehdr_part, m->parts[i]->type, m->parts[i]->fn);
		NTM_SEND_WRITE(buf, strlen(buf));
		NTM_SEND_WRITE(m->parts[i]->data, m->parts[i]->size);
	}
	NTM_SEND_LITERAL("\n\n");

ntm_send_done:
	(void)fflush(p);
	if ((e = pclose(p)) == 0 || e == 75)
		return 0;
	return 1;

ntm_send_error:
	(void)pclose(p);
	return -1;
}

char *
ntm_doinsets(const char *in, const NTM_INSET **insets)
{
	char obuf[65536];
	char sbuf[128];
	int i, k, y;

	if (insets == NULL)
		return nt_strdup(in);

	k = 0;
	for (i = 0; in[i] != '\0' && (size_t)k < sizeof(obuf) - 1; i++) {
ntm_doinsets_run:
		if (in[i] != '_') {
			obuf[k++] = in[i];
			continue;
		}

		/* try insets */
		for (y = 0; insets[y] != NULL; y++) {
			(void)snprintf(sbuf, sizeof(sbuf), "_%s_", insets[y]->tag);
			if (strncmp(in + i, sbuf, strlen(sbuf)) != 0)
				continue;

			/* found, substitute */
			(void)snprintf(obuf + k, sizeof(obuf) - k, "%s", insets[y]->sub);
			k += strlen(insets[y]->sub);
			i += strlen(sbuf);
			goto ntm_doinsets_run;
		}

		/* no inset found */
		obuf[k++] = in[i];
	}
	obuf[k] = '\0';

	return nt_strdup(obuf);
}

char *
ntm_dotpl(const char *tplfile, const NTM_INSET **insets)
{
	FILE *fp;
	char buf[65536];
	size_t len;

	if ((fp = fopen(tplfile, "r")) == NULL)
		return NULL;

	len = 0;
	while (len < sizeof(buf) && fread(buf + len, 1, 1, fp) == 1)
		len++;
	buf[len] = '\0';

	(void)fclose(fp);

	return ntm_doinsets(buf, insets);
}

NTM_INSET **
ntm_makeinsets(const char **tags, const char **subs)
{
	NTM_INSET **insets;
	int count, i;

	if (tags == NULL || subs == NULL)
		return NULL;

	count = 2;
	insets = nt_calloc((size_t)count, sizeof(NTM_INSET *));

	for (i = 0; tags[i] != NULL && subs[i] != NULL; i++) {
		if (i > count - 2) {
			count += 4;
			insets = nt_realloc(insets, count * sizeof(NTM_INSET *));
		}
		insets[i] = nt_calloc(1, sizeof(NTM_INSET));
		insets[i]->tag = nt_strdup(tags[i]);
		insets[i]->sub = nt_strdup(subs[i]);
	}
	insets[i] = NULL;

	return insets;
}

void
ntm_insets_free(NTM_INSET **insets)
{
	int i;

	for (i = 0; insets[i] != NULL; i++) {
		free(insets[i]->tag);
		free(insets[i]->sub);
		free(insets[i]);
	}
	free(insets);
}

NTM_PART **
ntm_makeparts(const char **fns, const char **rfns, const char **types)
{
	NTM_PART **parts;
	FILE *fp;
	char *buf;
	size_t len, bufsize;
	int count, i;

	if (fns == NULL || rfns == NULL || types == NULL)
		return NULL;

	count = 2;
	parts = nt_calloc((size_t)count, sizeof(NTM_PART *));

	for (i = 0; fns[i] != NULL && rfns[i] != NULL && types[i] != NULL; i++) {
		if (i > count - 2) {
			count += 4;
			parts = nt_realloc(parts, count * sizeof(NTM_PART *));
		}
		parts[i] = nt_calloc(1, sizeof(NTM_PART));
		parts[i]->fn = nt_strdup(fns[i]);
		parts[i]->type = nt_strdup(types[i]);
		parts[i+1] = NULL;

		if ((fp = fopen(rfns[i], "r")) == NULL)
			goto ntg_makeparts_error;

		bufsize = 65536; /* 64 KB */
		buf = nt_calloc(bufsize, sizeof(char));
		for (len = 0; len < NTM_PART_MAX_SIZE; len++) {
			if (len >= bufsize - 1) {
				bufsize += 65536;
				buf = nt_realloc(buf, bufsize * sizeof(char));
			}
			if (fread(buf + len, 1, 1, fp) != 1)
				break;
		}
		parts[i]->size = base64_encode_alloc(buf, len, &parts[i]->data);
		free(buf);
		(void)fclose(fp);
	}
	parts[i] = NULL;

	return parts;

ntg_makeparts_error:
	ntm_parts_free(parts);
	return NULL;
}

void
ntm_parts_free(NTM_PART **parts)
{
	int i;

	for (i = 0; parts[i] != NULL; i++) {
		free(parts[i]->fn);
		free(parts[i]->type);
		free(parts[i]->data);
		free(parts[i]);
	}
	free(parts);
}

int
ntm_tplsend(NTM_MAIL *mail, const char *tpl, const char **tags, const char **subs)
{
	NTM_INSET **insets;
	int e;

	if ((insets = ntm_makeinsets(tags, subs)) == NULL) {
		e = -1;
		return e;
	}

	mail->text = ntm_dotpl(tpl, (const NTM_INSET **)insets);
	e = ntm_send(mail);

	ntm_insets_free(insets);
	if (mail->text != NULL)
		free(mail->text);

	return e;
}


#ifdef TEST
int
main(void)
{
	NTM_MAIL m;
	NTM_INSET **insets;
	char *p;
	const char *tags[] = { "SUBST", "SUBST2", NULL };
	const char *subs[] = { "(subst)", "(subst2)", NULL };
	const char *fns[] = { "passwd", "fstab", "sh", NULL };
	const char *rfns[] = { "/etc/passwd", "/etc/fstab", "/bin/sh", NULL };
	const char *type[] = { "text/plain", "text/plain", "application/binary", NULL };

	(void)memset(&m, 0, sizeof(NTM_MAIL));
	m.from = "Test";
	m.to = nt_calloc(2, sizeof(char *));
	m.to[0] = "Foo";
	m.to[1] = NULL;
	m.subject = "Testmail";
	m.text = "Bar!";
	m.parts = nt_calloc(2, sizeof(NTM_PART *));
	m.parts[0] = nt_calloc(1, sizeof(NTM_PART));
	m.parts[0]->fn = "bibi.pdf";
	m.parts[0]->type = "application/pdf";
	m.parts[0]->data = "Text\nText\nText\nText\n";
	m.parts[0]->size = strlen(m.parts[0]->data);
	m.parts[1] = NULL;

	(void)fprintf(stderr, "(%d)\n", ntm_send(&m));
	free(m.parts[0]);
	free(m.parts);

	insets = ntm_makeinsets(tags, subs);
	p = ntm_dotpl("/tmp/fn0rd", (const NTM_INSET **)insets);
	(void)fprintf(stderr, "=[\n%s\n]=\n", p);
	ntm_insets_free(insets);
	if (p != NULL)
		free(p);

	m.parts = ntm_makeparts(fns, rfns, type);
	(void)fprintf(stderr, "(%d)\n", ntm_send(&m));
	ntm_parts_free(m.parts);
	m.parts = NULL;

	(void)fprintf(stderr, "(%d)\n", ntm_tplsend(&m, "/tmp/fn0rd", tags, subs));
	free(m.to);

	return EXIT_SUCCESS;
}
#endif
